@@ -1,19 +1,24 @@
+mod interlace_info;
 mod stream;
+pub(crate) mod transform;
 mod zlib;
 
 pub use self::stream::{DecodeOptions, Decoded, DecodingError, StreamingDecoder};
-use self::stream::{FormatErrorInner, CHUNCK_BUFFER_SIZE};
+use self::stream::{FormatErrorInner, CHUNK_BUFFER_SIZE};
+use self::transform::{create_transform_fn, TransformFn};
 
-use std::io::{BufRead, BufReader, Read, Write};
+use std::io::{BufRead, BufReader, ErrorKind, Read};
 use std::mem;
-use std::ops::Range;
 
+use crate::adam7::{self, Adam7Info};
 use crate::chunk;
 use crate::common::{
     BitDepth, BytesPerPixel, ColorType, Info, ParameterErrorKind, Transformations,
 };
 use crate::filter::{unfilter, FilterType};
-use crate::utils;
+
+pub use interlace_info::InterlaceInfo;
+use interlace_info::InterlaceInfoIter;
 
 /*
 pub enum InterlaceHandling {
@@ -59,6 +64,17 @@ pub struct Limits {
     pub bytes: usize,
 }
 
+impl Limits {
+    pub(crate) fn reserve_bytes(&mut self, bytes: usize) -> Result<(), DecodingError> {
+        if self.bytes >= bytes {
+            self.bytes -= bytes;
+            Ok(())
+        } else {
+            Err(DecodingError::LimitsExceeded)
+        }
+    }
+}
+
 impl Default for Limits {
     fn default() -> Limits {
         Limits {
@@ -72,8 +88,6 @@ pub struct Decoder<R: Read> {
     read_decoder: ReadDecoder<R>,
     /// Output transformations
     transform: Transformations,
-    /// Limits on resources the Decoder is allowed to use
-    limits: Limits,
 }
 
 /// A row of data with interlace information attached.
@@ -88,30 +102,11 @@ impl<'data> InterlacedRow<'data> {
         self.data
     }
 
-    pub fn interlace(&self) -> InterlaceInfo {
-        self.interlace
+    pub fn interlace(&self) -> &InterlaceInfo {
+        &self.interlace
     }
 }
 
-/// PNG (2003) specifies two interlace modes, but reserves future extensions.
-#[derive(Clone, Copy, Debug)]
-pub enum InterlaceInfo {
-    /// the null method means no interlacing
-    Null,
-    /// Adam7 derives its name from doing 7 passes over the image, only decoding a subset of all pixels in each pass.
-    /// The following table shows pictorially what parts of each 8x8 area of the image is found in each pass:
-    ///
-    /// 1 6 4 6 2 6 4 6
-    /// 7 7 7 7 7 7 7 7
-    /// 5 6 5 6 5 6 5 6
-    /// 7 7 7 7 7 7 7 7
-    /// 3 6 4 6 3 6 4 6
-    /// 7 7 7 7 7 7 7 7
-    /// 5 6 5 6 5 6 5 6
-    /// 7 7 7 7 7 7 7 7
-    Adam7 { pass: u8, line: u32, width: u32 },
-}
-
 /// A row of data without interlace information.
 #[derive(Clone, Copy, Debug)]
 pub struct Row<'data> {
@@ -132,27 +127,31 @@ impl<R: Read> Decoder<R> {
 
     /// Create a new decoder configuration with custom limits.
     pub fn new_with_limits(r: R, limits: Limits) -> Decoder<R> {
+        let mut decoder = StreamingDecoder::new();
+        decoder.limits = limits;
+
         Decoder {
             read_decoder: ReadDecoder {
-                reader: BufReader::with_capacity(CHUNCK_BUFFER_SIZE, r),
-                decoder: StreamingDecoder::new(),
+                reader: BufReader::with_capacity(CHUNK_BUFFER_SIZE, r),
+                decoder,
                 at_eof: false,
             },
             transform: Transformations::IDENTITY,
-            limits,
         }
     }
 
     /// Create a new decoder configuration with custom `DecodeOptions`.
     pub fn new_with_options(r: R, decode_options: DecodeOptions) -> Decoder<R> {
+        let mut decoder = StreamingDecoder::new_with_options(decode_options);
+        decoder.limits = Limits::default();
+
         Decoder {
             read_decoder: ReadDecoder {
-                reader: BufReader::with_capacity(CHUNCK_BUFFER_SIZE, r),
-                decoder: StreamingDecoder::new_with_options(decode_options),
+                reader: BufReader::with_capacity(CHUNK_BUFFER_SIZE, r),
+                decoder,
                 at_eof: false,
             },
             transform: Transformations::IDENTITY,
-            limits: Limits::default(),
         }
     }
 
@@ -180,21 +179,19 @@ impl<R: Read> Decoder<R> {
     /// assert!(decoder.read_info().is_ok());
     /// ```
     pub fn set_limits(&mut self, limits: Limits) {
-        self.limits = limits;
+        self.read_decoder.decoder.limits = limits;
     }
 
     /// Read the PNG header and return the information contained within.
     ///
     /// Most image metadata will not be read until `read_info` is called, so those fields will be
     /// None or empty.
-    pub fn read_header_info(&mut self) -> Result<&Info, DecodingError> {
+    pub fn read_header_info(&mut self) -> Result<&Info<'static>, DecodingError> {
         let mut buf = Vec::new();
         while self.read_decoder.info().is_none() {
             buf.clear();
             if self.read_decoder.decode_next(&mut buf)?.is_none() {
-                return Err(DecodingError::Format(
-                    FormatErrorInner::UnexpectedEof.into(),
-                ));
+                return Err(DecodingError::IoError(ErrorKind::UnexpectedEof.into()));
             }
         }
         Ok(self.read_decoder.info().unwrap())
@@ -210,12 +207,12 @@ impl<R: Read> Decoder<R> {
             subframe: SubframeInfo::not_yet_init(),
             fctl_read: 0,
             next_frame: SubframeIdx::Initial,
-            prev: Vec::new(),
-            current: Vec::new(),
-            scan_start: 0,
+            data_stream: Vec::new(),
+            prev_start: 0,
+            current_start: 0,
             transform: self.transform,
-            processed: Vec::new(),
-            limits: self.limits,
+            transform_fn: None,
+            scratch_buffer: Vec::new(),
         };
 
         // Check if the decoding buffer of a single raw line has a valid size.
@@ -264,6 +261,22 @@ impl<R: Read> Decoder<R> {
             .set_ignore_text_chunk(ignore_text_chunk);
     }
 
+    /// Set the decoder to ignore iccp chunks while parsing.
+    ///
+    /// eg.
+    /// ```
+    /// use std::fs::File;
+    /// use png::Decoder;
+    /// let mut decoder = Decoder::new(File::open("tests/iccp/broken_iccp.png").unwrap());
+    /// decoder.set_ignore_iccp_chunk(true);
+    /// assert!(decoder.read_info().is_ok());
+    /// ```
+    pub fn set_ignore_iccp_chunk(&mut self, ignore_iccp_chunk: bool) {
+        self.read_decoder
+            .decoder
+            .set_ignore_iccp_chunk(ignore_iccp_chunk);
+    }
+
     /// Set the decoder to ignore and not verify the Adler-32 checksum
     /// and CRC code.
     pub fn ignore_checksums(&mut self, ignore_checksums: bool) {
@@ -288,9 +301,7 @@ impl<R: Read> ReadDecoder<R> {
             let (consumed, result) = {
                 let buf = self.reader.fill_buf()?;
                 if buf.is_empty() {
-                    return Err(DecodingError::Format(
-                        FormatErrorInner::UnexpectedEof.into(),
-                    ));
+                    return Err(DecodingError::IoError(ErrorKind::UnexpectedEof.into()));
                 }
                 self.decoder.update(buf, image_data)?
             };
@@ -308,9 +319,7 @@ impl<R: Read> ReadDecoder<R> {
         while !self.at_eof {
             let buf = self.reader.fill_buf()?;
             if buf.is_empty() {
-                return Err(DecodingError::Format(
-                    FormatErrorInner::UnexpectedEof.into(),
-                ));
+                return Err(DecodingError::IoError(ErrorKind::UnexpectedEof.into()));
             }
             let (consumed, event) = self.decoder.update(buf, &mut vec![])?;
             self.reader.consume(consumed);
@@ -325,12 +334,10 @@ impl<R: Read> ReadDecoder<R> {
             }
         }
 
-        Err(DecodingError::Format(
-            FormatErrorInner::UnexpectedEof.into(),
-        ))
+        Err(DecodingError::IoError(ErrorKind::UnexpectedEof.into()))
     }
 
-    fn info(&self) -> Option<&Info> {
+    fn info(&self) -> Option<&Info<'static>> {
         self.decoder.info.as_ref()
     }
 }
@@ -347,18 +354,21 @@ pub struct Reader<R: Read> {
     /// control chunk. The IDAT image _may_ have such a chunk applying to it.
     fctl_read: u32,
     next_frame: SubframeIdx,
-    /// Previous raw line
-    prev: Vec<u8>,
-    /// Current raw line
-    current: Vec<u8>,
-    /// Start index of the current scan line.
-    scan_start: usize,
+    /// Vec containing the uncompressed image data currently being processed.
+    data_stream: Vec<u8>,
+    /// Index in `data_stream` where the previous row starts.
+    prev_start: usize,
+    /// Index in `data_stream` where the current row starts.
+    current_start: usize,
     /// Output transformations
     transform: Transformations,
-    /// Processed line
-    processed: Vec<u8>,
-    /// How resources we can spend (for example, on allocation).
-    limits: Limits,
+    /// Function that can transform decompressed, unfiltered rows into final output.
+    /// See the `transform.rs` module for more details.
+    transform_fn: Option<TransformFn>,
+    /// This buffer is only used so that `next_row` and `next_interlaced_row` can return reference
+    /// to a byte slice. In a future version of this library, this buffer will be removed and
+    /// `next_row` and `next_interlaced_row` will write directly into a user provided output buffer.
+    scratch_buffer: Vec<u8>,
 }
 
 /// The subframe specific information.
@@ -371,18 +381,13 @@ struct SubframeInfo {
     width: u32,
     height: u32,
     rowlen: usize,
-    interlace: InterlaceIter,
+    current_interlace_info: Option<InterlaceInfo>,
+    interlace_info_iter: InterlaceInfoIter,
     consumed_and_flushed: bool,
 }
 
-#[derive(Clone)]
-enum InterlaceIter {
-    None(Range<u32>),
-    Adam7(utils::Adam7Iterator),
-}
-
 /// Denote a frame as given by sequence numbers.
-#[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
+#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord)]
 enum SubframeIdx {
     /// The initial frame in an IDAT chunk without fcTL chunk applying to it.
     /// Note that this variant precedes `Some` as IDAT frames precede fdAT frames and all fdAT
@@ -410,7 +415,6 @@ impl<R: Read> Reader<R> {
                 Some(Decoded::ChunkBegin(_, chunk::IDAT))
                 | Some(Decoded::ChunkBegin(_, chunk::fdAT)) => break,
                 Some(Decoded::FrameControl(_)) => {
-                    self.subframe = SubframeInfo::new(self.info());
                     // The next frame is the one to which this chunk applies.
                     self.next_frame = SubframeIdx::Some(self.fctl_read);
                     // TODO: what about overflow here? That would imply there are more fctl chunks
@@ -435,16 +439,15 @@ impl<R: Read> Reader<R> {
             .ok_or(DecodingError::Format(FormatErrorInner::MissingIhdr.into()))?;
         self.bpp = info.bpp_in_prediction();
         self.subframe = SubframeInfo::new(info);
+        self.data_stream.clear();
+        self.current_start = 0;
+        self.prev_start = 0;
 
         // Allocate output buffer.
-        let buflen = self
-            .line_size(self.subframe.width)
-            .filter(|&x| x <= self.limits.bytes)
-            .ok_or(DecodingError::LimitsExceeded)?;
-        self.processed.resize(buflen, 0u8);
+        let buflen = self.output_line_size(self.subframe.width);
+        self.decoder.decoder.limits.reserve_bytes(buflen)?;
 
-        self.prev.clear();
-        self.prev.resize(self.subframe.rowlen, 0);
+        self.prev_start = self.current_start;
 
         Ok(())
     }
@@ -452,7 +455,7 @@ impl<R: Read> Reader<R> {
     /// Get information on the image.
     ///
     /// The structure will change as new frames of an animated image are decoded.
-    pub fn info(&self) -> &Info {
+    pub fn info(&self) -> &Info<'static> {
         self.decoder.info().unwrap()
     }
 
@@ -470,17 +473,13 @@ impl<R: Read> Reader<R> {
     /// Output lines will be written in row-major, packed matrix with width and height of the read
     /// frame (or subframe), all samples are in big endian byte order where this matters.
     pub fn next_frame(&mut self, buf: &mut [u8]) -> Result<OutputInfo, DecodingError> {
-        let subframe_idx = match self.decoder.info().unwrap().frame_control() {
-            None => SubframeIdx::Initial,
-            Some(_) => SubframeIdx::Some(self.fctl_read - 1),
-        };
-
         if self.next_frame == SubframeIdx::End {
             return Err(DecodingError::Parameter(
                 ParameterErrorKind::PolledAfterEndOfImage.into(),
             ));
-        } else if self.next_frame != subframe_idx {
-            // Advance until we've read the info / fcTL for this frame.
+        } else if self.subframe.consumed_and_flushed {
+            // Advance until the next `fdAT`
+            // (along the way we should encounter the fcTL for this frame).
             self.read_until_image_data()?;
         }
 
@@ -503,42 +502,47 @@ impl<R: Read> Reader<R> {
             line_size: self.output_line_size(self.subframe.width),
         };
 
-        self.current.clear();
-        self.scan_start = 0;
-        let width = self.info().width;
         if self.info().interlaced {
+            let stride = self.output_line_size(self.info().width);
+            let samples = color_type.samples() as u8;
+            let bits_pp = samples * (bit_depth as u8);
             while let Some(InterlacedRow {
                 data: row,
                 interlace,
                 ..
             }) = self.next_interlaced_row()?
             {
-                let (line, pass) = match interlace {
-                    InterlaceInfo::Adam7 { line, pass, .. } => (line, pass),
-                    InterlaceInfo::Null => unreachable!("expected interlace information"),
-                };
-                let samples = color_type.samples() as u8;
-                utils::expand_pass(buf, width, row, pass, line, samples * (bit_depth as u8));
+                // `unwrap` won't panic, because we checked `self.info().interlaced` above.
+                let adam7info = interlace.get_adam7_info().unwrap();
+                adam7::expand_pass(buf, stride, row, adam7info, bits_pp);
             }
         } else {
-            let mut len = 0;
-            while let Some(Row { data: row, .. }) = self.next_row()? {
-                len += (&mut buf[len..]).write(row)?;
+            let current_interlace_info = self.subframe.current_interlace_info.as_ref();
+            let already_done_rows = current_interlace_info
+                .map(|info| info.line_number())
+                .unwrap_or(self.subframe.height);
+
+            for row in buf
+                .chunks_exact_mut(output_info.line_size)
+                .take(self.subframe.height as usize)
+                .skip(already_done_rows as usize)
+            {
+                self.next_interlaced_row_impl(self.subframe.rowlen, row)?;
             }
         }
 
         // Advance over the rest of data for this (sub-)frame.
         if !self.subframe.consumed_and_flushed {
             self.decoder.finish_decoding()?;
+            self.subframe.consumed_and_flushed = true;
         }
 
         // Advance our state to expect the next frame.
-        let past_end_subframe = match self.info().animation_control() {
-            // a non-APNG has no subframes
-            None => 0,
-            // otherwise the count is the past-the-end index. It can not be 0 per spec.
-            Some(ac) => ac.num_frames,
-        };
+        let past_end_subframe = self
+            .info()
+            .animation_control()
+            .map(|ac| ac.num_frames)
+            .unwrap_or(0);
         self.next_frame = match self.next_frame {
             SubframeIdx::End => unreachable!("Next frame called when already at image end"),
             // Reached the end of non-animated image.
@@ -562,75 +566,79 @@ impl<R: Read> Reader<R> {
 
     /// Returns the next processed row of the image
     pub fn next_interlaced_row(&mut self) -> Result<Option<InterlacedRow>, DecodingError> {
-        self.next_interlaced_row_impl()
+        let interlace = match self.subframe.current_interlace_info.as_ref() {
+            None => return Ok(None),
+            Some(interlace) => *interlace,
+        };
+        if interlace.line_number() == 0 {
+            self.prev_start = self.current_start;
+        }
+        let rowlen = match interlace {
+            InterlaceInfo::Null(_) => self.subframe.rowlen,
+            InterlaceInfo::Adam7(Adam7Info { width, .. }) => {
+                self.info().raw_row_length_from_width(width)
+            }
+        };
+        let width = match interlace {
+            InterlaceInfo::Adam7(Adam7Info { width, .. }) => width,
+            InterlaceInfo::Null(_) => self.subframe.width,
+        };
+        let output_line_size = self.output_line_size(width);
+
+        // TODO: change the interface of `next_interlaced_row` to take an output buffer instead of
+        // making us return a reference to a buffer that we own.
+        let mut output_buffer = mem::take(&mut self.scratch_buffer);
+        output_buffer.resize(output_line_size, 0u8);
+        let ret = self.next_interlaced_row_impl(rowlen, &mut output_buffer);
+        self.scratch_buffer = output_buffer;
+        ret?;
+
+        Ok(Some(InterlacedRow {
+            data: &self.scratch_buffer[..output_line_size],
+            interlace,
+        }))
+    }
+
+    /// Read the rest of the image and chunks and finish up, including text chunks or others
+    /// This will discard the rest of the image if the image is not read already with [`Reader::next_frame`], [`Reader::next_row`] or [`Reader::next_interlaced_row`]
+    pub fn finish(&mut self) -> Result<(), DecodingError> {
+        self.next_frame = SubframeIdx::End;
+        self.data_stream.clear();
+        self.current_start = 0;
+        self.prev_start = 0;
+        loop {
+            let mut buf = Vec::new();
+            let state = self.decoder.decode_next(&mut buf)?;
+
+            if state.is_none() {
+                break;
+            }
+        }
+
+        Ok(())
     }
 
     /// Fetch the next interlaced row and filter it according to our own transformations.
-    fn next_interlaced_row_impl(&mut self) -> Result<Option<InterlacedRow>, DecodingError> {
-        use crate::common::ColorType::*;
-        let transform = self.transform;
+    fn next_interlaced_row_impl(
+        &mut self,
+        rowlen: usize,
+        output_buffer: &mut [u8],
+    ) -> Result<(), DecodingError> {
+        self.next_raw_interlaced_row(rowlen)?;
+        assert_eq!(self.current_start - self.prev_start, rowlen - 1);
+        let row = &self.data_stream[self.prev_start..self.current_start];
 
-        let (rowlen, interlace) = match self.next_pass() {
-            Some((rowlen, interlace)) => (rowlen, interlace),
-            None => return Ok(None),
-        };
-
-        if transform == Transformations::IDENTITY {
-            return Ok(Some(InterlacedRow {
-                data: self.next_raw_interlaced_row(rowlen)?,
-                interlace,
-            }));
-        }
-
-        // swap buffer to circumvent borrow issues
-        let mut buffer = mem::take(&mut self.processed);
-        (&mut buffer[..]).write_all(self.next_raw_interlaced_row(rowlen)?)?;
-        self.processed = buffer;
-
-        let (color_type, bit_depth, trns) = {
-            let info = self.info();
-            (info.color_type, info.bit_depth as u8, info.trns.is_some())
-        };
-        let output_buffer = if let InterlaceInfo::Adam7 { width, .. } = interlace {
-            let width = self
-                .line_size(width)
-                .expect("Adam7 interlaced rows are shorter than the buffer.");
-            &mut self.processed[..width]
-        } else {
-            &mut *self.processed
-        };
-
-        let mut len = output_buffer.len();
-        if transform.contains(Transformations::EXPAND) {
-            match color_type {
-                Indexed => expand_paletted(output_buffer, self.decoder.info().unwrap())?,
-                Grayscale | GrayscaleAlpha if bit_depth < 8 => {
-                    expand_gray_u8(output_buffer, self.decoder.info().unwrap())
-                }
-                Grayscale | Rgb if trns => {
-                    let channels = color_type.samples();
-                    let trns = self.decoder.info().unwrap().trns.as_ref().unwrap();
-                    if bit_depth == 8 {
-                        utils::expand_trns_line(output_buffer, trns, channels);
-                    } else {
-                        utils::expand_trns_line16(output_buffer, trns, channels);
-                    }
-                }
-                _ => (),
+        // Apply transformations and write resulting data to buffer.
+        let transform_fn = {
+            if self.transform_fn.is_none() {
+                self.transform_fn = Some(create_transform_fn(self.info(), self.transform)?);
             }
-        }
+            self.transform_fn.as_deref().unwrap()
+        };
+        transform_fn(row, output_buffer, self.info());
 
-        if bit_depth == 16 && transform.intersects(Transformations::STRIP_16) {
-            len /= 2;
-            for i in 0..len {
-                output_buffer[i] = output_buffer[2 * i];
-            }
-        }
-
-        Ok(Some(InterlacedRow {
-            data: &output_buffer[..len],
-            interlace,
-        }))
+        self.subframe.current_interlace_info = self.subframe.interlace_info_iter.next();
+        Ok(())
     }
 
     /// Returns the color type and the number of bits per sample
@@ -644,21 +652,27 @@ impl<R: Read> Reader<R> {
         } else {
             let bits = match info.bit_depth as u8 {
                 16 if t.intersects(Transformations::STRIP_16) => 8,
-                n if n < 8 && t.contains(Transformations::EXPAND) => 8,
+                n if n < 8
+                    && (t.contains(Transformations::EXPAND)
+                        || t.contains(Transformations::ALPHA)) =>
+                {
+                    8
+                }
                 n => n,
             };
-            let color_type = if t.contains(Transformations::EXPAND) {
-                let has_trns = info.trns.is_some();
-                match info.color_type {
-                    Grayscale if has_trns => GrayscaleAlpha,
-                    Rgb if has_trns => Rgba,
-                    Indexed if has_trns => Rgba,
-                    Indexed => Rgb,
-                    ct => ct,
-                }
-            } else {
-                info.color_type
-            };
+            let color_type =
+                if t.contains(Transformations::EXPAND) || t.contains(Transformations::ALPHA) {
+                    let has_trns = info.trns.is_some() || t.contains(Transformations::ALPHA);
+                    match info.color_type {
+                        Grayscale if has_trns => GrayscaleAlpha,
+                        Rgb if has_trns => Rgba,
+                        Indexed if has_trns => Rgba,
+                        Indexed => Rgb,
+                        ct => ct,
+                    }
+                } else {
+                    info.color_type
+                };
             (color_type, BitDepth::from_u8(bits).unwrap())
         }
     }
@@ -677,58 +691,12 @@ impl<R: Read> Reader<R> {
         color.raw_row_length_from_width(depth, width) - 1
     }
 
-    /// Returns the number of bytes required to decode a deinterlaced row.
-    fn line_size(&self, width: u32) -> Option<usize> {
-        use crate::common::ColorType::*;
-        let t = self.transform;
-        let info = self.info();
-        let trns = info.trns.is_some();
-
-        let expanded = if info.bit_depth == BitDepth::Sixteen {
-            BitDepth::Sixteen
-        } else {
-            BitDepth::Eight
-        };
-        // The color type and depth representing the decoded line
-        // TODO 16 bit
-        let (color, depth) = match info.color_type {
-            Indexed if trns && t.contains(Transformations::EXPAND) => (Rgba, expanded),
-            Indexed if t.contains(Transformations::EXPAND) => (Rgb, expanded),
-            Rgb if trns && t.contains(Transformations::EXPAND) => (Rgba, expanded),
-            Grayscale if trns && t.contains(Transformations::EXPAND) => (GrayscaleAlpha, expanded),
-            Grayscale if t.contains(Transformations::EXPAND) => (Grayscale, expanded),
-            GrayscaleAlpha if t.contains(Transformations::EXPAND) => (GrayscaleAlpha, expanded),
-            other => (other, info.bit_depth),
-        };
-
-        // Without the filter method byte
-        color.checked_raw_row_length(depth, width).map(|n| n - 1)
-    }
-
-    fn next_pass(&mut self) -> Option<(usize, InterlaceInfo)> {
-        match self.subframe.interlace {
-            InterlaceIter::Adam7(ref mut adam7) => {
-                let last_pass = adam7.current_pass();
-                let (pass, line, width) = adam7.next()?;
-                let rowlen = self.info().raw_row_length_from_width(width);
-                if last_pass != pass {
-                    self.prev.clear();
-                    self.prev.resize(rowlen, 0u8);
-                }
-                Some((rowlen, InterlaceInfo::Adam7 { pass, line, width }))
-            }
-            InterlaceIter::None(ref mut height) => {
-                let _ = height.next()?;
-                Some((self.subframe.rowlen, InterlaceInfo::Null))
-            }
-        }
-    }
-
-    /// Returns the next raw scanline of the image interlace pass.
+    /// Write the next raw interlaced row into `self.prev`.
+    ///
     /// The scanline is filtered against the previous scanline according to the specification.
-    fn next_raw_interlaced_row(&mut self, rowlen: usize) -> Result<&[u8], DecodingError> {
+    fn next_raw_interlaced_row(&mut self, rowlen: usize) -> Result<(), DecodingError> {
         // Read image data until we have at least one full row (but possibly more than one).
-        while self.current.len() - self.scan_start < rowlen {
+        while self.data_stream.len() - self.current_start < rowlen {
             if self.subframe.consumed_and_flushed {
                 return Err(DecodingError::Format(
                     FormatErrorInner::NoMoreImageData.into(),
@@ -736,44 +704,39 @@ impl<R: Read> Reader<R> {
             }
 
             // Clear the current buffer before appending more data.
-            if self.scan_start > 0 {
-                self.current.drain(..self.scan_start).for_each(drop);
-                self.scan_start = 0;
+            if self.prev_start > 0 {
+                self.data_stream.copy_within(self.prev_start.., 0);
+                self.data_stream
+                    .truncate(self.data_stream.len() - self.prev_start);
+                self.current_start -= self.prev_start;
+                self.prev_start = 0;
             }
 
-            match self.decoder.decode_next(&mut self.current)? {
-                Some(Decoded::ImageData) => {}
-                Some(Decoded::ImageDataFlushed) => {
-                    self.subframe.consumed_and_flushed = true;
-                }
-                None => {
-                    return Err(DecodingError::Format(
-                        if self.current.is_empty() {
-                            FormatErrorInner::NoMoreImageData
-                        } else {
-                            FormatErrorInner::UnexpectedEndOfChunk
-                        }
-                        .into(),
-                    ));
-                }
+            match self.decoder.decode_next(&mut self.data_stream)? {
+                Some(Decoded::ImageData) => (),
+                Some(Decoded::ImageDataFlushed) => self.subframe.consumed_and_flushed = true,
                 _ => (),
             }
         }
 
         // Get a reference to the current row and point scan_start to the next one.
-        let row = &mut self.current[self.scan_start..];
-        self.scan_start += rowlen;
+        let (prev, row) = self.data_stream.split_at_mut(self.current_start);
 
         // Unfilter the row.
         let filter = FilterType::from_u8(row[0]).ok_or(DecodingError::Format(
             FormatErrorInner::UnknownFilterMethod(row[0]).into(),
         ))?;
-        unfilter(filter, self.bpp, &self.prev[1..rowlen], &mut row[1..rowlen]);
+        unfilter(
+            filter,
+            self.bpp,
+            &prev[self.prev_start..],
+            &mut row[1..rowlen],
+        );
 
-        // Save the current row for the next pass.
-        self.prev[..rowlen].copy_from_slice(&row[..rowlen]);
+        self.prev_start = self.current_start + 1;
+        self.current_start += rowlen;
 
-        Ok(&self.prev[1..rowlen])
+        Ok(())
     }
 }
 
@@ -783,7 +746,8 @@ impl SubframeInfo {
             width: 0,
             height: 0,
             rowlen: 0,
-            interlace: InterlaceIter::None(0..0),
+            current_interlace_info: None,
+            interlace_info_iter: InterlaceInfoIter::empty(),
             consumed_and_flushed: false,
         }
     }
@@ -797,159 +761,15 @@ impl SubframeInfo {
             (info.width, info.height)
         };
 
-        let interlace = if info.interlaced {
-            InterlaceIter::Adam7(utils::Adam7Iterator::new(width, height))
-        } else {
-            InterlaceIter::None(0..height)
-        };
-
+        let mut interlace_info_iter = InterlaceInfoIter::new(width, height, info.interlaced);
+        let current_interlace_info = interlace_info_iter.next();
         SubframeInfo {
             width,
             height,
             rowlen: info.raw_row_length_from_width(width),
-            interlace,
+            current_interlace_info,
+            interlace_info_iter,
             consumed_and_flushed: false,
         }
     }
 }
-
-fn expand_paletted(buffer: &mut [u8], info: &Info) -> Result<(), DecodingError> {
-    if let Some(palette) = info.palette.as_ref() {
-        if let BitDepth::Sixteen = info.bit_depth {
-            // This should have been caught earlier but let's check again. Can't hurt.
-            Err(DecodingError::Format(
-                FormatErrorInner::InvalidColorBitDepth {
-                    color_type: ColorType::Indexed,
-                    bit_depth: BitDepth::Sixteen,
-                }
-                .into(),
-            ))
-        } else {
-            let black = [0, 0, 0];
-            if let Some(ref trns) = info.trns {
-                // > The tRNS chunk shall not contain more alpha values than there are palette
-                // entries, but a tRNS chunk may contain fewer values than there are palette
-                // entries. In this case, the alpha value for all remaining palette entries is
-                // assumed to be 255.
-                //
-                // It seems, accepted reading is to fully *ignore* an invalid tRNS as if it were
-                // completely empty / all pixels are non-transparent.
-                let trns = if trns.len() <= palette.len() / 3 {
-                    trns.as_ref()
-                } else {
-                    &[]
-                };
-
-                utils::unpack_bits(buffer, 4, info.bit_depth as u8, |i, chunk| {
-                    let (rgb, a) = (
-                        palette
-                            .get(3 * i as usize..3 * i as usize + 3)
-                            .unwrap_or(&black),
-                        *trns.get(i as usize).unwrap_or(&0xFF),
-                    );
-                    chunk[0] = rgb[0];
-                    chunk[1] = rgb[1];
-                    chunk[2] = rgb[2];
-                    chunk[3] = a;
-                });
-            } else {
-                utils::unpack_bits(buffer, 3, info.bit_depth as u8, |i, chunk| {
-                    let rgb = palette
-                        .get(3 * i as usize..3 * i as usize + 3)
-                        .unwrap_or(&black);
-                    chunk[0] = rgb[0];
-                    chunk[1] = rgb[1];
-                    chunk[2] = rgb[2];
-                })
-            }
-            Ok(())
-        }
-    } else {
-        Err(DecodingError::Format(
-            FormatErrorInner::PaletteRequired.into(),
-        ))
-    }
-}
-
-fn expand_gray_u8(buffer: &mut [u8], info: &Info) {
-    let rescale = true;
-    let scaling_factor = if rescale {
-        (255) / ((1u16 << info.bit_depth as u8) - 1) as u8
-    } else {
-        1
-    };
-    if let Some(ref trns) = info.trns {
-        utils::unpack_bits(buffer, 2, info.bit_depth as u8, |pixel, chunk| {
-            if pixel == trns[0] {
-                chunk[1] = 0
-            } else {
-                chunk[1] = 0xFF
-            }
-            chunk[0] = pixel * scaling_factor
-        })
-    } else {
-        utils::unpack_bits(buffer, 1, info.bit_depth as u8, |val, chunk| {
-            chunk[0] = val * scaling_factor
-        })
-    }
-}
-
-#[cfg(test)]
-mod tests {
-    use super::Decoder;
-    use std::io::{BufRead, Read, Result};
-    use std::mem::discriminant;
-
-    /// A reader that reads at most `n` bytes.
-    struct SmalBuf<R: BufRead> {
-        inner: R,
-        cap: usize,
-    }
-
-    impl<R: BufRead> SmalBuf<R> {
-        fn new(inner: R, cap: usize) -> Self {
-            SmalBuf { inner, cap }
-        }
-    }
-
-    impl<R: BufRead> Read for SmalBuf<R> {
-        fn read(&mut self, buf: &mut [u8]) -> Result<usize> {
-            let len = buf.len().min(self.cap);
-            self.inner.read(&mut buf[..len])
-        }
-    }
-
-    impl<R: BufRead> BufRead for SmalBuf<R> {
-        fn fill_buf(&mut self) -> Result<&[u8]> {
-            let buf = self.inner.fill_buf()?;
-            let len = buf.len().min(self.cap);
-            Ok(&buf[..len])
-        }
-
-        fn consume(&mut self, amt: usize) {
-            assert!(amt <= self.cap);
-            self.inner.consume(amt)
-        }
-    }
-
-    #[test]
-    fn no_data_dup_on_finish() {
-        const IMG: &[u8] = include_bytes!(concat!(
-            env!("CARGO_MANIFEST_DIR"),
-            "/tests/bugfixes/x_issue#214.png"
-        ));
-
-        let mut normal = Decoder::new(IMG).read_info().unwrap();
-
-        let mut buffer = vec![0; normal.output_buffer_size()];
-        let normal = normal.next_frame(&mut buffer).unwrap_err();
-
-        let smal = Decoder::new(SmalBuf::new(IMG, 1))
-            .read_info()
-            .unwrap()
-            .next_frame(&mut buffer)
-            .unwrap_err();
-
-        assert_eq!(discriminant(&normal), discriminant(&smal));
-    }
-}
