@@ -1,7 +1,4 @@
-extern crate crc32fast;
-
-use std::convert::{From, TryInto};
-use std::default::Default;
+use std::convert::TryInto;
 use std::error;
 use std::fmt;
 use std::io;
@@ -20,7 +17,7 @@ use crate::traits::ReadBytesExt;
 use crate::Limits;
 
 /// TODO check if these size are reasonable
-pub const CHUNCK_BUFFER_SIZE: usize = 32 * 1024;
+pub const CHUNK_BUFFER_SIZE: usize = 32 * 1024;
 
 /// Determines if checksum checks should be disabled globally.
 ///
@@ -111,6 +108,11 @@ pub enum Decoded {
 #[derive(Debug)]
 pub enum DecodingError {
     /// An error in IO of the underlying reader.
+    ///
+    /// Note that some IO errors may be recoverable - decoding may be retried after the
+    /// error is resolved.  For example, decoding from a slow stream of data (e.g. decoding from a
+    /// network stream) may occasionally result in [std::io::ErrorKind::UnexpectedEof] kind of
+    /// error, but decoding can resume when more data becomes available.
     IoError(io::Error),
     /// The input image was not a valid PNG.
     ///
@@ -163,10 +165,6 @@ pub(crate) enum FormatErrorInner {
     },
     /// Not a PNG, the magic signature is missing.
     InvalidSignature,
-    /// End of file, within a chunk event.
-    UnexpectedEof,
-    /// End of file, while expecting more image data.
-    UnexpectedEndOfChunk,
     // Errors of chunk level ordering, missing etc.
     /// Ihdr must occur.
     MissingIhdr,
@@ -230,7 +228,7 @@ pub(crate) enum FormatErrorInner {
     /// The subframe is not in bounds of the image.
     /// TODO: fields with relevant data.
     BadSubFrameBounds {},
-    // Errors specific to the IDAT/fDAT chunks.
+    // Errors specific to the IDAT/fdAT chunks.
     /// The compression of the data stream was faulty.
     CorruptFlateStream {
         err: fdeflate::DecompressionError,
@@ -241,6 +239,21 @@ pub(crate) enum FormatErrorInner {
     BadTextEncoding(TextDecodingError),
     /// fdAT shorter than 4 bytes
     FdatShorterThanFourBytes,
+    /// "11.2.4 IDAT Image data" section of the PNG spec says: There may be multiple IDAT chunks;
+    /// if so, they shall appear consecutively with no other intervening chunks.
+    /// `UnexpectedRestartOfDataChunkSequence{kind: IDAT}` indicates that there were "intervening
+    /// chunks".
+    ///
+    /// The APNG spec doesn't directly describe an error similar to `CantInterleaveIdatChunks`,
+    /// but we require that a new sequence of consecutive `fdAT` chunks cannot appear unless we've
+    /// seen an `fcTL` chunk.
+    UnexpectedRestartOfDataChunkSequence {
+        kind: ChunkType,
+    },
+    /// Failure to parse a chunk, because the chunk didn't contain enough bytes.
+    ChunkTooShort {
+        kind: ChunkType,
+    },
 }
 
 impl error::Error for DecodingError {
@@ -280,7 +293,7 @@ impl fmt::Display for FormatError {
             ),
             MissingIhdr => write!(fmt, "IHDR chunk missing"),
             MissingFctl => write!(fmt, "fcTL chunk missing before fdAT chunk."),
-            MissingImageData => write!(fmt, "IDAT or fDAT chunk is missing."),
+            MissingImageData => write!(fmt, "IDAT or fdAT chunk is missing."),
             ChunkBeforeIhdr { kind } => write!(fmt, "{:?} chunk appeared before IHDR chunk", kind),
             AfterIdat { kind } => write!(fmt, "Chunk {:?} is invalid after IDAT chunk.", kind),
             AfterPlte { kind } => write!(fmt, "Chunk {:?} is invalid after PLTE chunk.", kind),
@@ -315,7 +328,7 @@ impl fmt::Display for FormatError {
                 "Transparency chunk found for color type {:?}.",
                 color_type
             ),
-            InvalidBitDepth(nr) => write!(fmt, "Invalid dispose operation {}.", nr),
+            InvalidBitDepth(nr) => write!(fmt, "Invalid bit depth {}.", nr),
             InvalidColorType(nr) => write!(fmt, "Invalid color type {}.", nr),
             InvalidDisposeOp(nr) => write!(fmt, "Invalid dispose op {}.", nr),
             InvalidBlendOp(nr) => write!(fmt, "Invalid blend op {}.", nr),
@@ -326,9 +339,10 @@ impl fmt::Display for FormatError {
             UnknownInterlaceMethod(nr) => write!(fmt, "Unknown interlace method {}.", nr),
             BadSubFrameBounds {} => write!(fmt, "Sub frame is out-of-bounds."),
             InvalidSignature => write!(fmt, "Invalid PNG signature."),
-            UnexpectedEof => write!(fmt, "Unexpected end of data before image end."),
-            UnexpectedEndOfChunk => write!(fmt, "Unexpected end of data within a chunk."),
-            NoMoreImageData => write!(fmt, "IDAT or fDAT chunk is has not enough data for image."),
+            NoMoreImageData => write!(
+                fmt,
+                "IDAT or fDAT chunk does not have enough data for image."
+            ),
             CorruptFlateStream { err } => {
                 write!(fmt, "Corrupt deflate stream. ")?;
                 write!(fmt, "{:?}", err)
@@ -363,6 +377,12 @@ impl fmt::Display for FormatError {
                 }
             }
             FdatShorterThanFourBytes => write!(fmt, "fdAT chunk shorter than 4 bytes"),
+            UnexpectedRestartOfDataChunkSequence { kind } => {
+                write!(fmt, "Unexpected restart of {:?} chunk sequence", kind)
+            }
+            ChunkTooShort { kind } => {
+                write!(fmt, "Chunk is too short: {:?}", kind)
+            }
         }
     }
 }
@@ -408,6 +428,7 @@ pub struct DecodeOptions {
     ignore_adler32: bool,
     ignore_crc: bool,
     ignore_text_chunk: bool,
+    ignore_iccp_chunk: bool,
     skip_ancillary_crc_failures: bool,
 }
 
@@ -417,6 +438,7 @@ impl Default for DecodeOptions {
             ignore_adler32: true,
             ignore_crc: false,
             ignore_text_chunk: false,
+            ignore_iccp_chunk: false,
             skip_ancillary_crc_failures: true,
         }
     }
@@ -451,6 +473,13 @@ impl DecodeOptions {
         self.ignore_text_chunk = ignore_text_chunk;
     }
 
+    /// Ignore ICCP chunks while decoding.
+    ///
+    /// Defaults to `false`.
+    pub fn set_ignore_iccp_chunk(&mut self, ignore_iccp_chunk: bool) {
+        self.ignore_iccp_chunk = ignore_iccp_chunk;
+    }
+
     /// Ignore ancillary chunks if CRC fails
     ///
     /// Defaults to `true`
@@ -476,6 +505,15 @@ pub struct StreamingDecoder {
     /// Whether we have already seen a start of an IDAT chunk.  (Used to validate chunk ordering -
     /// some chunk types can only appear before or after an IDAT chunk.)
     have_idat: bool,
+    /// Whether we are ready for a start of an `IDAT` chunk sequence.  Initially `true` and set to
+    /// `false` when the first sequence of consecutive `IDAT` chunks ends.
+    ready_for_idat_chunks: bool,
+    /// Whether we are ready for a start of an `fdAT` chunk sequence.  Initially `false`.  Set to
+    /// `true` after encountering an `fcTL` chunk. Set to `false` when a sequence of consecutive
+    /// `fdAT` chunks ends.
+    ready_for_fdat_chunks: bool,
+    /// Whether we have already seen an iCCP chunk. Used to prevent parsing of duplicate iCCP chunks.
+    have_iccp: bool,
     decode_options: DecodeOptions,
     pub(crate) limits: Limits,
 }
@@ -514,6 +552,9 @@ impl StreamingDecoder {
             info: None,
             current_seq_no: None,
             have_idat: false,
+            have_iccp: false,
+            ready_for_idat_chunks: true,
+            ready_for_fdat_chunks: false,
             decode_options,
             limits: Limits { bytes: usize::MAX },
         }
@@ -540,6 +581,10 @@ impl StreamingDecoder {
         self.decode_options.set_ignore_text_chunk(ignore_text_chunk);
     }
 
+    pub fn set_ignore_iccp_chunk(&mut self, ignore_iccp_chunk: bool) {
+        self.decode_options.set_ignore_iccp_chunk(ignore_iccp_chunk);
+    }
+
     /// Return whether the decoder is set to ignore the Adler-32 checksum.
     pub fn ignore_adler32(&self) -> bool {
         self.inflater.ignore_adler32()
@@ -761,6 +806,8 @@ impl StreamingDecoder {
                     self.current_chunk.type_ = type_str;
                     self.inflater.finish_compressed_chunks(image_data)?;
                     self.inflater.reset();
+                    self.ready_for_idat_chunks = false;
+                    self.ready_for_fdat_chunks = false;
                     self.state = Some(State::U32 {
                         kind,
                         bytes,
@@ -768,15 +815,16 @@ impl StreamingDecoder {
                     });
                     return Ok(Decoded::ImageDataFlushed);
                 }
-                self.current_chunk.type_ = type_str;
-                if !self.decode_options.ignore_crc {
-                    self.current_chunk.crc.reset();
-                    self.current_chunk.crc.update(&type_str.0);
-                }
-                self.current_chunk.remaining = length;
-                self.current_chunk.raw_bytes.clear();
                 self.state = match type_str {
                     chunk::fdAT => {
+                        if !self.ready_for_fdat_chunks {
+                            return Err(DecodingError::Format(
+                                FormatErrorInner::UnexpectedRestartOfDataChunkSequence {
+                                    kind: chunk::fdAT,
+                                }
+                                .into(),
+                            ));
+                        }
                         if length < 4 {
                             return Err(DecodingError::Format(
                                 FormatErrorInner::FdatShorterThanFourBytes.into(),
@@ -785,11 +833,26 @@ impl StreamingDecoder {
                         Some(State::new_u32(U32ValueKind::ApngSequenceNumber))
                     }
                     IDAT => {
+                        if !self.ready_for_idat_chunks {
+                            return Err(DecodingError::Format(
+                                FormatErrorInner::UnexpectedRestartOfDataChunkSequence {
+                                    kind: IDAT,
+                                }
+                                .into(),
+                            ));
+                        }
                         self.have_idat = true;
                         Some(State::ImageData(type_str))
                     }
                     _ => Some(State::ReadChunkData(type_str)),
                 };
+                self.current_chunk.type_ = type_str;
+                if !self.decode_options.ignore_crc {
+                    self.current_chunk.crc.reset();
+                    self.current_chunk.crc.update(&type_str.0);
+                }
+                self.current_chunk.remaining = length;
+                self.current_chunk.raw_bytes.clear();
                 Ok(Decoded::ChunkBegin(length, type_str))
             }
             U32ValueKind::Crc(type_str) => {
@@ -888,16 +951,29 @@ impl StreamingDecoder {
             chunk::fcTL => self.parse_fctl(),
             chunk::cHRM => self.parse_chrm(),
             chunk::sRGB => self.parse_srgb(),
-            chunk::iCCP => self.parse_iccp(),
+            chunk::iCCP if !self.decode_options.ignore_iccp_chunk => self.parse_iccp(),
             chunk::tEXt if !self.decode_options.ignore_text_chunk => self.parse_text(),
             chunk::zTXt if !self.decode_options.ignore_text_chunk => self.parse_ztxt(),
             chunk::iTXt if !self.decode_options.ignore_text_chunk => self.parse_itxt(),
             _ => Ok(Decoded::PartialChunk(type_str)),
         };
 
-        if parse_result.is_err() {
+        let parse_result = parse_result.map_err(|e| {
             self.state = None;
-        }
+            match e {
+                // `parse_chunk` is invoked after gathering **all** bytes of a chunk, so
+                // `UnexpectedEof` from something like `read_be` is permanent and indicates an
+                // invalid PNG that should be represented as a `FormatError`, rather than as a
+                // (potentially recoverable) `IoError` / `UnexpectedEof`.
+                DecodingError::IoError(e) if e.kind() == std::io::ErrorKind::UnexpectedEof => {
+                    let fmt_err: FormatError =
+                        FormatErrorInner::ChunkTooShort { kind: type_str }.into();
+                    fmt_err.into()
+                }
+                e => e,
+            }
+        });
+
         parse_result
     }
 
@@ -930,6 +1006,7 @@ impl StreamingDecoder {
             0
         });
         self.inflater.reset();
+        self.ready_for_fdat_chunks = true;
         let fc = FrameControl {
             sequence_number: next_seq_no,
             width: buf.read_be()?,
@@ -1190,12 +1267,11 @@ impl StreamingDecoder {
     }
 
     fn parse_iccp(&mut self) -> Result<Decoded, DecodingError> {
-        let info = self.info.as_mut().unwrap();
         if self.have_idat {
             Err(DecodingError::Format(
                 FormatErrorInner::AfterIdat { kind: chunk::iCCP }.into(),
             ))
-        } else if info.icc_profile.is_some() {
+        } else if self.have_iccp {
             // We have already encountered an iCCP chunk before.
             //
             // Section "4.2.2.4. iCCP Embedded ICC profile" of the spec says:
@@ -1209,46 +1285,53 @@ impl StreamingDecoder {
             //     (treating them as a benign error).
             Ok(Decoded::Nothing)
         } else {
-            let mut buf = &self.current_chunk.raw_bytes[..];
-
-            // read profile name
-            let _: u8 = buf.read_be()?;
-            for _ in 1..80 {
-                let raw: u8 = buf.read_be()?;
-                if raw == 0 {
-                    break;
-                }
-            }
-
-            match buf.read_be()? {
-                // compression method
-                0u8 => (),
-                n => {
-                    return Err(DecodingError::Format(
-                        FormatErrorInner::UnknownCompressionMethod(n).into(),
-                    ))
-                }
-            }
-
-            match fdeflate::decompress_to_vec_bounded(buf, self.limits.bytes) {
-                Ok(profile) => {
-                    self.limits.reserve_bytes(profile.len())?;
-                    info.icc_profile = Some(Cow::Owned(profile));
-                }
-                Err(fdeflate::BoundedDecompressionError::DecompressionError { inner: err }) => {
-                    return Err(DecodingError::Format(
-                        FormatErrorInner::CorruptFlateStream { err }.into(),
-                    ))
-                }
-                Err(fdeflate::BoundedDecompressionError::OutputTooLarge { .. }) => {
-                    return Err(DecodingError::LimitsExceeded);
-                }
-            }
-
+            self.have_iccp = true;
+            let _ = self.parse_iccp_raw();
             Ok(Decoded::Nothing)
         }
     }
 
+    fn parse_iccp_raw(&mut self) -> Result<(), DecodingError> {
+        let info = self.info.as_mut().unwrap();
+        let mut buf = &self.current_chunk.raw_bytes[..];
+
+        // read profile name
+        let _: u8 = buf.read_be()?;
+        for _ in 1..80 {
+            let raw: u8 = buf.read_be()?;
+            if raw == 0 {
+                break;
+            }
+        }
+
+        match buf.read_be()? {
+            // compression method
+            0u8 => (),
+            n => {
+                return Err(DecodingError::Format(
+                    FormatErrorInner::UnknownCompressionMethod(n).into(),
+                ))
+            }
+        }
+
+        match fdeflate::decompress_to_vec_bounded(buf, self.limits.bytes) {
+            Ok(profile) => {
+                self.limits.reserve_bytes(profile.len())?;
+                info.icc_profile = Some(Cow::Owned(profile));
+            }
+            Err(fdeflate::BoundedDecompressionError::DecompressionError { inner: err }) => {
+                return Err(DecodingError::Format(
+                    FormatErrorInner::CorruptFlateStream { err }.into(),
+                ))
+            }
+            Err(fdeflate::BoundedDecompressionError::OutputTooLarge { .. }) => {
+                return Err(DecodingError::LimitsExceeded);
+            }
+        }
+
+        Ok(())
+    }
+
     fn parse_ihdr(&mut self) -> Result<Decoded, DecodingError> {
         if self.info.is_some() {
             return Err(DecodingError::Format(
@@ -1444,6 +1527,12 @@ impl StreamingDecoder {
 
 impl Info<'_> {
     fn validate(&self, fc: &FrameControl) -> Result<(), DecodingError> {
+        if fc.width == 0 || fc.height == 0 {
+            return Err(DecodingError::Format(
+                FormatErrorInner::InvalidDimensions.into(),
+            ));
+        }
+
         // Validate mathematically: fc.width + fc.x_offset <= self.width
         let in_x_bounds = Some(fc.width) <= self.width.checked_sub(fc.x_offset);
         // Validate mathematically: fc.height + fc.y_offset <= self.height
@@ -1472,7 +1561,7 @@ impl Default for ChunkState {
             type_: ChunkType([0; 4]),
             crc: Crc32::new(),
             remaining: 0,
-            raw_bytes: Vec::with_capacity(CHUNCK_BUFFER_SIZE),
+            raw_bytes: Vec::with_capacity(CHUNK_BUFFER_SIZE),
         }
     }
 }
@@ -1482,10 +1571,12 @@ mod tests {
     use super::ScaledFloat;
     use super::SourceChromaticities;
     use crate::test_utils::*;
-    use crate::{Decoder, DecodingError};
+    use crate::{Decoder, DecodingError, Reader};
     use byteorder::WriteBytesExt;
+    use std::cell::RefCell;
     use std::fs::File;
-    use std::io::Write;
+    use std::io::{ErrorKind, Read, Write};
+    use std::rc::Rc;
 
     #[test]
     fn image_gamma() -> Result<(), ()> {
@@ -1734,6 +1825,15 @@ mod tests {
         assert_eq!(4070462061, crc32fast::hash(&icc_profile));
     }
 
+    #[test]
+    fn test_png_with_broken_iccp() {
+        let decoder = crate::Decoder::new(File::open("tests/iccp/broken_iccp.png").unwrap());
+        assert!(decoder.read_info().is_ok());
+        let mut decoder = crate::Decoder::new(File::open("tests/iccp/broken_iccp.png").unwrap());
+        decoder.set_ignore_iccp_chunk(true);
+        assert!(decoder.read_info().is_ok());
+    }
+
     /// Writes an acTL chunk.
     /// See https://wiki.mozilla.org/APNG_Specification#.60acTL.60:_The_Animation_Control_Chunk
     fn write_actl(w: &mut impl Write, animation: &crate::AnimationControl) {
@@ -1903,7 +2003,7 @@ mod tests {
                 panic!("No fcTL (2nd frame)");
             };
             // The sequence number is taken from the `fcTL` chunk that comes before the two `fdAT`
-            // chunks.  Note that sequence numbers inside `fdAT` chunks are not publically exposed
+            // chunks.  Note that sequence numbers inside `fdAT` chunks are not publicly exposed
             // (but they are still checked when decoding to verify that they are sequential).
             assert_eq!(frame_control.sequence_number, 1);
         }
@@ -1939,4 +2039,174 @@ mod tests {
         reader.next_frame(&mut buf).unwrap();
         assert_eq!(3093270825, crc32fast::hash(&buf));
     }
+
+    /// `StremingInput` can be used by tests to simulate a streaming input
+    /// (e.g. a slow http response, where all bytes are not immediately available).
+    #[derive(Clone)]
+    struct StreamingInput(Rc<RefCell<StreamingInputState>>);
+
+    struct StreamingInputState {
+        full_input: Vec<u8>,
+        current_pos: usize,
+        available_len: usize,
+    }
+
+    impl StreamingInput {
+        fn new(full_input: Vec<u8>) -> Self {
+            Self(Rc::new(RefCell::new(StreamingInputState {
+                full_input,
+                current_pos: 0,
+                available_len: 0,
+            })))
+        }
+
+        fn with_noncompressed_png(width: u32, idat_size: usize) -> Self {
+            let mut png = Vec::new();
+            write_noncompressed_png(&mut png, width, idat_size);
+            Self::new(png)
+        }
+
+        fn expose_next_byte(&self) {
+            let mut state = self.0.borrow_mut();
+            assert!(state.available_len < state.full_input.len());
+            state.available_len += 1;
+        }
+
+        fn stream_input_until_reader_is_available(&self) -> Reader<StreamingInput> {
+            loop {
+                self.0.borrow_mut().current_pos = 0;
+                match Decoder::new(self.clone()).read_info() {
+                    Ok(reader) => {
+                        break reader;
+                    }
+                    Err(DecodingError::IoError(e)) if e.kind() == ErrorKind::UnexpectedEof => {
+                        self.expose_next_byte();
+                    }
+                    _ => panic!("Unexpected error"),
+                }
+            }
+        }
+
+        fn decode_full_input<F, R>(&self, f: F) -> R
+        where
+            F: FnOnce(Reader<&[u8]>) -> R,
+        {
+            let state = self.0.borrow();
+            let decoder = Decoder::new(state.full_input.as_slice());
+            f(decoder.read_info().unwrap())
+        }
+    }
+
+    impl Read for StreamingInput {
+        fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
+            let mut state = self.0.borrow_mut();
+            let mut available_bytes = &state.full_input[state.current_pos..state.available_len];
+            let number_of_read_bytes = available_bytes.read(buf)?;
+            state.current_pos += number_of_read_bytes;
+            assert!(state.current_pos <= state.available_len);
+            Ok(number_of_read_bytes)
+        }
+    }
+
+    /// Test resuming/retrying `Reader.next_frame` after `UnexpectedEof`.
+    #[test]
+    fn test_streaming_input_and_decoding_via_next_frame() {
+        const WIDTH: u32 = 16;
+        const IDAT_SIZE: usize = 512;
+        let streaming_input = StreamingInput::with_noncompressed_png(WIDTH, IDAT_SIZE);
+
+        let (whole_output_info, decoded_from_whole_input) =
+            streaming_input.decode_full_input(|mut r| {
+                let mut buf = vec![0; r.output_buffer_size()];
+                let output_info = r.next_frame(&mut buf).unwrap();
+                (output_info, buf)
+            });
+
+        let mut png_reader = streaming_input.stream_input_until_reader_is_available();
+        let mut decoded_from_streaming_input = vec![0; png_reader.output_buffer_size()];
+        let streaming_output_info = loop {
+            match png_reader.next_frame(decoded_from_streaming_input.as_mut_slice()) {
+                Ok(output_info) => break output_info,
+                Err(DecodingError::IoError(e)) if e.kind() == ErrorKind::UnexpectedEof => {
+                    streaming_input.expose_next_byte()
+                }
+                e => panic!("Unexpected error: {:?}", e),
+            }
+        };
+        assert_eq!(whole_output_info, streaming_output_info);
+        assert_eq!(
+            crc32fast::hash(&decoded_from_whole_input),
+            crc32fast::hash(&decoded_from_streaming_input)
+        );
+    }
+
+    /// Test resuming/retrying `Reader.next_row` after `UnexpectedEof`.
+    #[test]
+    fn test_streaming_input_and_decoding_via_next_row() {
+        const WIDTH: u32 = 16;
+        const IDAT_SIZE: usize = 512;
+        let streaming_input = StreamingInput::with_noncompressed_png(WIDTH, IDAT_SIZE);
+
+        let decoded_from_whole_input = streaming_input.decode_full_input(|mut r| {
+            let mut buf = vec![0; r.output_buffer_size()];
+            r.next_frame(&mut buf).unwrap();
+            buf
+        });
+
+        let mut png_reader = streaming_input.stream_input_until_reader_is_available();
+        let mut decoded_from_streaming_input = Vec::new();
+        loop {
+            match png_reader.next_row() {
+                Ok(None) => break,
+                Ok(Some(row)) => decoded_from_streaming_input.extend_from_slice(row.data()),
+                Err(DecodingError::IoError(e)) if e.kind() == ErrorKind::UnexpectedEof => {
+                    streaming_input.expose_next_byte()
+                }
+                e => panic!("Unexpected error: {:?}", e),
+            }
+        }
+        assert_eq!(
+            crc32fast::hash(&decoded_from_whole_input),
+            crc32fast::hash(&decoded_from_streaming_input)
+        );
+    }
+
+    /// Test resuming/retrying `Decoder.read_header_info` after `UnexpectedEof`.
+    #[test]
+    fn test_streaming_input_and_reading_header_info() {
+        const WIDTH: u32 = 16;
+        const IDAT_SIZE: usize = 512;
+        let streaming_input = StreamingInput::with_noncompressed_png(WIDTH, IDAT_SIZE);
+
+        let info_from_whole_input = streaming_input.decode_full_input(|r| r.info().clone());
+
+        let mut decoder = Decoder::new(streaming_input.clone());
+        let info_from_streaming_input = loop {
+            match decoder.read_header_info() {
+                Ok(info) => break info.clone(),
+                Err(DecodingError::IoError(e)) if e.kind() == ErrorKind::UnexpectedEof => {
+                    streaming_input.expose_next_byte()
+                }
+                e => panic!("Unexpected error: {:?}", e),
+            }
+        };
+
+        assert_eq!(info_from_whole_input.width, info_from_streaming_input.width);
+        assert_eq!(
+            info_from_whole_input.height,
+            info_from_streaming_input.height
+        );
+        assert_eq!(
+            info_from_whole_input.bit_depth,
+            info_from_streaming_input.bit_depth
+        );
+        assert_eq!(
+            info_from_whole_input.color_type,
+            info_from_streaming_input.color_type
+        );
+        assert_eq!(
+            info_from_whole_input.interlaced,
+            info_from_streaming_input.interlaced
+        );
+    }
 }
