@@ -4,17 +4,16 @@
 //!
 //! ## The decoder
 //!
-//! The most important types for decoding purposes are [`Decoder`](struct.Decoder.html) and
-//! [`Reader`](struct.Reader.html). They both wrap a `std::io::Read`.
-//! `Decoder` serves as a builder for `Reader`. Calling `Decoder::read_info` reads from the `Read` until the
+//! The most important types for decoding purposes are [`Decoder`] and
+//! [`Reader`]. They both wrap a [`std::io::Read`].
+//! `Decoder` serves as a builder for `Reader`. Calling [`Decoder::read_info`] reads from the `Read` until the
 //! image data is reached.
 //!
 //! ### Using the decoder
 //! ```
 //! use std::fs::File;
 //! // The decoder is a build for reader and can be used to set various decoding options
-//! // via `Transformations`. The default output transformation is `Transformations::EXPAND
-//! // | Transformations::STRIP_ALPHA`.
+//! // via `Transformations`. The default output transformation is `Transformations::IDENTITY`.
 //! let decoder = png::Decoder::new(File::open("tests/pngsuite/basi0g01.png").unwrap());
 //! let mut reader = decoder.read_info().unwrap();
 //! // Allocate the output buffer.
@@ -43,7 +42,6 @@
 //! let mut encoder = png::Encoder::new(w, 2, 1); // Width is 2 pixels and height is 1.
 //! encoder.set_color(png::ColorType::Rgba);
 //! encoder.set_depth(png::BitDepth::Eight);
-//! encoder.set_trns(vec!(0xFFu8, 0xFFu8, 0xFFu8, 0xFFu8));
 //! encoder.set_source_gamma(png::ScaledFloat::from_scaled(45455)); // 1.0 / 2.2, scaled by 100000
 //! encoder.set_source_gamma(png::ScaledFloat::new(1.0 / 2.2));     // 1.0 / 2.2, unscaled, but rounded
 //! let source_chromaticities = png::SourceChromaticities::new(     // Using unscaled instantiation here
@@ -60,23 +58,31 @@
 //! ```
 //!
 
+#![cfg_attr(feature = "unstable", feature(portable_simd))]
 #![forbid(unsafe_code)]
 
-#[macro_use]
-extern crate bitflags;
-
+mod adam7;
 pub mod chunk;
 mod common;
 mod decoder;
 mod encoder;
 mod filter;
 mod srgb;
+pub mod text_metadata;
 mod traits;
-mod utils;
 
+pub use crate::adam7::expand_pass as expand_interlaced_row;
+pub use crate::adam7::Adam7Info;
 pub use crate::common::*;
 pub use crate::decoder::{
-    Decoded, Decoder, DecodingError, Limits, OutputInfo, Reader, StreamingDecoder,
+    DecodeOptions, Decoded, Decoder, DecodingError, InterlaceInfo, InterlacedRow, Limits,
+    OutputInfo, Reader, StreamingDecoder,
 };
 pub use crate::encoder::{Encoder, EncodingError, StreamWriter, Writer};
 pub use crate::filter::{AdaptiveFilterType, FilterType};
+
+#[cfg(test)]
+pub(crate) mod test_utils;
+
+#[cfg(feature = "benchmarks")]
+pub mod benchable_apis;
